"""
Handles Creation and Update of Sphinx Documentation

If no docs directory exists: Will create an initial docs directory
and populate Sphinx configuration with default settings.

Also generates API documentation for the source code in the
configured source code directory.

NOTE: Users will still need to update the generated .rst
      files to provide further documentation. This only
      generates API
"""
import os
import getpass
from io import open

PROJECT_NAME = "cirrus"
DOCUMENTATION_DIR = os.path.dirname(os.path.realpath(__file__))
SOURCE_CODE_DIR = os.path.abspath(DOCUMENTATION_DIR + "/../")


def main():
    """
    Entry point of script, builds commands and executes
    """
    commands = _build_commands()

    if not os.path.exists(DOCUMENTATION_DIR + "/source"):
        os.makedirs(DOCUMENTATION_DIR + "/source")
        os.system(commands["create_initial_docs_command"])
        _create_custom_cfg_section_in_sphinx_conf()
        # _make_version_and_release_equal()
        _manually_append_source_path_to_sphinx_config()

    os.system(commands["retrieve_source_api_command"])
    os.system(commands["build_docs_command"])


def _build_commands():
    commands = dict()

    commands["create_initial_docs_command"] = (
        "sphinx-quickstart -p {PROJECT} -a {AUTHOR} -v {VERSION} --sep --makefile -q "
        "--ext-autodoc --ext-doctest --ext-coverage --ext-imgmath --ext-todo "
        "--ext-intersphinx --ext-viewcode "
        "--extensions sphinx.ext.autodoc,sphinx.ext.napoleon {DOCS_DIR}".format(
            PROJECT=PROJECT_NAME,
            AUTHOR=getpass.getuser(),
            VERSION=1.0,
            DOCS_DIR=DOCUMENTATION_DIR,
        )
    )

    commands[
        "retrieve_source_api_command"
    ] = "sphinx-apidoc -o {DOCS_DIR}/source {SOURCE_DIR}".format(
        DOCS_DIR=DOCUMENTATION_DIR, SOURCE_DIR=SOURCE_CODE_DIR
    )

    commands[
        "build_docs_command"
    ] = "sphinx-build -b html {DOCS_DIR}/source {DOCS_DIR}/build".format(
        DOCS_DIR=DOCUMENTATION_DIR
    )

    return commands


def _manually_append_source_path_to_sphinx_config():
    """
    Adds the source directory to the Python system path in the Sphinx config
    file so Sphinx knows where to access the source files.

    TODO FIXME
    NOTE: This appends new paths to the beginning of the path but doesn't
          remove previous paths added... This could cause issues in the
          future (and/or a long list of sys.path.insert's)
    """
    with open(
        DOCUMENTATION_DIR + "/source/conf.py", "a+", encoding="UTF-8"
    ) as sphinx_config:
        sphinx_config_abs_location = os.path.realpath(
            os.path.dirname(sphinx_config.name)
        )
        source_relative_to_sphinx_config = os.path.relpath(
            SOURCE_CODE_DIR, sphinx_config_abs_location
        )
        command_to_add_path = (
            'sys.path.insert(0, os.path.abspath("'
            + source_relative_to_sphinx_config
            + '"))'
        )

        sphinx_config.seek(0)

        if command_to_add_path not in sphinx_config.read():
            sphinx_config.write("\n" + command_to_add_path)


def _create_custom_cfg_section_in_sphinx_conf():
    with open(
        DOCUMENTATION_DIR + "/source/conf.py", "a+", encoding="UTF-8"
    ) as sphinx_config:
        sphinx_config.write("\n# CUSTOM CONFIG FROM DOCS CREATION SCRIPT #\n")
        sphinx_config.write("import os\n")
        sphinx_config.write("import sys\n")


def _make_version_and_release_equal():
    with open(
        DOCUMENTATION_DIR + "/source/conf.py", "a", encoding="UTF-8"
    ) as sphinx_config:
        sphinx_config.write("\nrelease = version\n")


if __name__ == "__main__":
    main()



"""
Google services for interacting with APIs.

See README for details on different ways to interact with Google's API(s)
"""
from googleapiclient.discovery import build
from cirrus.config import config


class GoogleService(object):
    """
    Generic Google servicing using Method 1 (Google's google-api-python-client)
    """

    def __init__(self, service_name, version, scopes, creds):
        """
        Create an object that can be used to build a service to interact
        with Google's APIs. This holds the necessary information and
        credentials to create a service.

        Args:
            service_name (str): Google service name
            version (str): Google service version
            scopes (List(str)): List of permission scopes to use when accessing API
            creds (google.oauth2.service_account.Credentials): SA creds
        """
        self.service_name = service_name
        self.version = version
        self.scopes = scopes
        self.creds = creds.with_scopes(scopes)

    def use_delegated_credentials(self, user_to_become):
        """
        Use current credentials to become another user.
        This allows service accounts with domain-wide delegation
        to immitate a specific user in their domain.

        Args:
            user_to_become (str): Email of user to become
        """
        delegated_credentials = self.creds.with_subject(user_to_become)
        self.creds = delegated_credentials

    def build_service(self):
        """
        Combines service, version, and creds to give a resource that
        can directly talk to Google APIs.

        See information here about Google's library:
        https://developers.google.com/api-client-library/python/start/get_started#building_and_calling_a_service

        Returns:
            googleapiclient.discovery.Resource: Google Resource to interact with
            API
        """
        return build(self.service_name, self.version, credentials=self.creds)


class GoogleAdminService(GoogleService):
    """
    Admin service is using Method 1 (Google's google-api-python-client)
    For Cloud Platform API's, Google recommends using the Google Cloud Client Library for Python

    Attributes:
        SCOPES (List(str)): Scopes required for permission to do group management
    """

    SCOPES = [
        "https://www.googleapis.com/auth/admin.directory.group",
        "https://www.googleapis.com/auth/admin.directory.group.readonly",
        "https://www.googleapis.com/auth/admin.directory.group.member",
        "https://www.googleapis.com/auth/admin.directory.group.member.readonly",
        "https://www.googleapis.com/auth/admin.directory.user.security",
    ]

    def __init__(self, creds):
        """
        Create the Google Admin Directory Service
        """
        super(GoogleAdminService, self).__init__(
            "admin", "directory_v1", self.SCOPES, creds=creds
        )
        self.use_delegated_credentials(config.GOOGLE_CLOUD_IDENTITY_ADMIN_EMAIL)



from cirrus.google_cloud.iam import (
    GooglePolicyMember,
    GooglePolicyRole,
    GooglePolicyBinding,
    GooglePolicy,
)


def test_create_google_policy():
    """
    Test if initialization of policy classes correctly sets up
    .roles and .members attributes
    :return:
    """

    # set up role and member objects
    role1 = GooglePolicyRole("roles/admin")
    mem11 = GooglePolicyMember("user", "mem11@gmail.com")
    mem12 = GooglePolicyMember("serviceAccount", "mem12@iam.gserviceaccount.com")
    mem13 = GooglePolicyMember("user", "mem13@gmail.com")

    role2 = GooglePolicyRole("roles/read")
    mem21 = GooglePolicyMember("user", "mem21@gmail.com")
    mem22 = GooglePolicyMember("serviceAccount", "mem22@iam.gserviceaccount.com")
    mem23 = GooglePolicyMember("user", "mem23@gmail.com")

    # group members for adding to policy binding
    # member23 is added to mems1 to make sure it isn't
    # double counted in overall policy member count"""
    mems1 = [mem11, mem12, mem13, mem23]
    mems2 = [mem21, mem22, mem23]

    # bind role to list of members
    bind1 = GooglePolicyBinding(role1, mems1)
    bind2 = GooglePolicyBinding(role2, mems2)

    for mem in mems1:
        # check that each member was given the proper role
        # mem23 should have both roles
        assert role1 in mem.roles
        # check that each binding has all the members it should
        assert mem in bind1.members
        # check that each role has the members it should
        assert mem in role1.members
    for mem in mems2:
        assert role2 in mem.roles
        assert mem in bind2.members
        assert mem in role2.members

    # create policy from list of bindings
    policy = GooglePolicy([bind1, bind2])

    # check number of members and roles
    # members should be 6 as mem23 is not double counted
    assert len(policy.members) == 6
    assert len(policy.roles) == 2

    # check that the policy has all the members and roles
    for mem in mems1 + mems2:
        assert mem in policy.members
        assert mem.roles.issubset(policy.roles)


def test_create_google_policy_from_json():
    """
    Test that google policy object is created correctly
    from json and has proper attributes
    """
    json = {
        "bindings": [
            {
                "role": "roles/admin",
                "members": [
                    "user:mem11@gmail.com",
                    "serviceAccount:mem12@iam.gserviceaccount.com",
                    "user:mem13@gmail.com",
                    "user:mem23@gmail.com",
                ],
            },
            {
                "role": "roles/read",
                "members": [
                    "user:mem21@gmail.com",
                    "serviceAccount:mem22@iam.gserviceaccount.com",
                    "user:mem23@gmail.com",
                ],
            },
        ]
    }

    # create policy from json
    policy = GooglePolicy.from_json(json)

    # set up sub-classes
    # these are just for comparing against the objects that
    # get created as a part of policy creation from json
    role1 = GooglePolicyRole("roles/admin")
    mem11 = GooglePolicyMember("user", "mem11@gmail.com")
    mem12 = GooglePolicyMember("serviceAccount", "mem12@iam.gserviceaccount.com")
    mem13 = GooglePolicyMember("user", "mem13@gmail.com")

    role2 = GooglePolicyRole("roles/read")
    mem21 = GooglePolicyMember("user", "mem21@gmail.com")
    mem22 = GooglePolicyMember("serviceAccount", "mem22@iam.gserviceaccount.com")
    mem23 = GooglePolicyMember("user", "mem23@gmail.com")

    mems1 = [mem11, mem12, mem13, mem23]
    mems2 = [mem21, mem22, mem23]

    GooglePolicyBinding(role1, mems1)
    GooglePolicyBinding(role2, mems2)

    # check policy has correct number of members and roles
    assert len(policy.members) == 6
    assert len(policy.roles) == 2

    # check policy has correct members and roles
    for mem in mems1 + mems2:
        assert mem in policy.members
        assert mem.roles.issubset(policy.roles)



from subprocess import check_output

from setuptools import setup


def get_version():
    # https://github.com/uc-cdis/dictionaryutils/pull/37#discussion_r257898408
    try:
        tag = check_output(
            ["git", "describe", "--tags", "--abbrev=0", "--match=[0-9]*"]
        )
        return tag.decode("utf-8").strip("\n")
    except Exception:
        raise RuntimeError(
            "The version number cannot be extracted from git tag in this source "
            "distribution; please either download the source from PyPI, or check out "
            "from GitHub and make sure that the git CLI is available."
        )


setup(
    name="gen3cirrus",
    version='1.3.0',
    install_requires=[
        "backoff~=1.6",
        "cdislogging",
        "oauth2client<4.0dev,>=2.0.0",
        "google-cloud-storage~=1.10",
        "google-api-python-client==1.11.0",
        "google-auth~=1.4",
        "google-auth-httplib2>=0.0.3",
    ],
    dependency_links=[
        "git+https://git@github.com/uc-cdis/cdislogging.git@master#egg=cdislogging"
    ],
    include_package_data=True,
    packages=["cirrus", "cirrus.google_cloud"],
)
